'''
===================
rpm.py
===================
Rock physics models
@author: Alessandro Amato del Monte
'''
import numpy as np


def critpor(K0, G0, phi, phic=0.4):
    """
    Critical porosity, Nur et al. (1991, 1995) written by Alessandro Amato
    del Monte (2015) from Mavko et al., Rock Physics Handbook, p.353.

    Args:
        K0, G0: float or array_like
            Mineral bulk & shear modulus in GPa
        phi: float or array_like
            porosity
        phic: float or array_like
            critical porosity (default 0.4)
    
    Returns:
        Tuple: K_DRY, G_DRY: dry rock bulk & shear modulus in GPa
    """
    K_DRY  = K0 * (1-phi/phic)
    G_DRY  = G0 * (1-phi/phic)
    return K_DRY, G_DRY


def hertzmindlin(K0, G0, phi, phic=0.4, Cn=8.6, P=10, f=1):
    '''
    Hertz-Mindlin model
    written by Alessandro Amato del Monte (2015)
    from Mavko et al., Rock Physics Handbook, p.246
    
    Args:
        K0, G0: mineral bulk & shear modulus in GPa
        phi: porosity
        phic: critical porosity (default 0.4)
        Cn: coordination nnumber (default 8.6)
        P: confining pressure in MPa (default 10)
        f: shear modulus correction factor, f=1 for dry pack with perfect adhesion between particles and f=0 for dry frictionless pack.
    
    Returns:
        Tuple: K_DRY, G_DRY: dry rock bulk & shear modulus in GPa
    '''
    P /= 1e3 # converts pressure in same units as solid moduli (GPa)
    PR0=(3*K0-2*G0)/(6*K0+2*G0) # poisson's ratio of mineral mixture
    K_HM = (P*(Cn**2*(1-phic)**2*G0**2) / (18*np.pi**2*(1-PR0)**2))**(1/3)
    G_HM = ((2+3*f-PR0*(1+3*f))/(5*(2-PR0))) * ((P*(3*Cn**2*(1-phic)**2*G0**2)/(2*np.pi**2*(1-PR0)**2)))**(1/3)
    return K_HM, G_HM

def softsand(K0, G0, phi, phic=0.4, Cn=8.6, P=10, f=1):
    '''
    Soft-sand (uncemented) model
    written by Alessandro Amato del Monte (2015)
    from Mavko et al., Rock Physics Handbook, p.258
    
    Args:
        K0, G0: mineral bulk & shear modulus in GPa
        phi: porosity
        phic: critical porosity (default 0.4)
        Cn: coordination nnumber (default 8.6)
        P: confining pressure in MPa (default 10)
        f: shear modulus correction factor, f=1 for dry pack with perfect adhesion between particles and f=0 for dry frictionless pack
    
    Returns:
        Tuple: K_DRY, G_DRY: dry rock bulk & shear modulus in GPa
    '''
    K_HM, G_HM = hertzmindlin(K0, G0, phi, phic, Cn, P, f)
    K_DRY =-4/3*G_HM + (((phi/phic)/(K_HM+4/3*G_HM)) + ((1-phi/phic)/(K0+4/3*G_HM)))**-1
    tmp = G_HM/6*((9*K_HM+8*G_HM) / (K_HM+2*G_HM))
    G_DRY = -tmp + ((phi/phic)/(G_HM+tmp) + ((1-phi/phic)/(G0+tmp)))**-1
    return K_DRY, G_DRY

def stiffsand(K0, G0, phi, phic=0.4, Cn=8.6, P=10, f=1):
    '''
    Stiff-sand model
    written by Alessandro Amato del Monte (2015)
    from Mavko et al., Rock Physics Handbook, p.260
    
    Args:
        K0, G0: mineral bulk & shear modulus in GPa
        phi: porosity
        phic: critical porosity (default 0.4)
        Cn: coordination nnumber (default 8.6)
        P: confining pressure in MPa (default 10)
        f: shear modulus correction factor, f=1 for dry pack with perfect adhesion between particles and f=0 for dry frictionless pack
    
    Returns
        Tuple: K_DRY, G_DRY: dry rock bulk & shear modulus in GPa
    '''
    K_HM, G_HM = hertzmindlin(K0, G0, phi, phic, Cn, P, f)
    K_DRY = -4/3*G0 + (((phi/phic)/(K_HM+4/3*G0)) + ((1-phi/phic)/(K0+4/3*G0)))**-1
    tmp = G0/6*((9*K0+8*G0) / (K0+2*G0))
    G_DRY = -tmp + ((phi/phic)/(G_HM+tmp) + ((1-phi/phic)/(G0+tmp)))**-1
    return K_DRY, G_DRY

def contactcement(K0, G0, phi, phic=0.4, Cn=8.6, Kc=37, Gc=45, scheme=2):
    '''
    Contact cement (cemented sand) model, Dvorkin-Nur (1996)
    written by Alessandro Amato del Monte (2015)
    from Mavko et al., Rock Physics Handbook, p.255
    
    Args:
        K0, G0: mineral bulk & shear modulus in GPa
        phi: porosity
        phic: critical porosity (default 0.4)
        Cn: coordination nnumber (default 8.6)
        Kc, Gc: cement bulk & shear modulus in GPa (default 37, 45 i.e. quartz)
        scheme: 1=cement deposited at grain contacts, 2=in uniform layer around grains (default 2)
    Returns:
        Tuple: K_DRY, G_DRY: dry rock bulk & shear modulus in GPa
    '''
    PR0=(3*K0-2*G0)/(6*K0+2*G0)
    PRc = (3*Kc-2*Gc)/(6*Kc+2*Gc)
    if scheme == 1: # scheme 1: cement deposited at grain contacts
        alpha = ((phic-phi)/(3*Cn*(1-phic))) ** (1/4)
    else: # scheme 2: cement evenly deposited on grain surface
        alpha = ((2*(phic-phi))/(3*(1-phic)))**(1/2)
    LambdaN = (2*Gc*(1-PR0)*(1-PRc)) / (np.pi*G0*(1-2*PRc))
    N1 = -0.024153*LambdaN**-1.3646
    N2 = 0.20405*LambdaN**-0.89008
    N3 = 0.00024649*LambdaN**-1.9864
    Sn = N1*alpha**2 + N2*alpha + N3
    LambdaT = Gc/(np.pi*G0)
    T1 = -10**-2*(2.26*PR0**2+2.07*PR0+2.3)*LambdaT**(0.079*PR0**2+0.1754*PR0-1.342)
    T2 = (0.0573*PR0**2+0.0937*PR0+0.202)*LambdaT**(0.0274*PR0**2+0.0529*PR0-0.8765)
    T3 = 10**-4*(9.654*PR0**2+4.945*PR0+3.1)*LambdaT**(0.01867*PR0**2+0.4011*PR0-1.8186)
    St = T1*alpha**2 + T2*alpha + T3
    K_DRY = 1/6*Cn*(1-phic)*(Kc+(4/3)*Gc)*Sn
    G_DRY = 3/5*K_DRY+3/20*Cn*(1-phic)*Gc*St
    return K_DRY, G_DRY



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Find and run tests.
"""
import unittest
import os


def run_tests():
    tests = []
    for root, _, _ in os.walk('bruges'):
        if 'test' in root:
            tests.append(
                unittest.TestLoader().discover(root,
                                               pattern='*test.py'))
    suite = unittest.TestSuite(tests)
    return(suite)


if __name__ == '__main__':
    suite = run_tests()
    unittest.TextTestRunner(verbosity=2).run(suite)



# -*- coding: utf-8 -*-
"""
Tests.
"""
import unittest
from numpy import linspace, sin, pi, amax

from bruges.attribute import energy


class EnergyTest(unittest.TestCase):

    def setUp(self):
        """
        Makes a simple sin wave with 1 amplitude to use as test data.
        """
        self.n_samples = 1001
        duration = 1.0
        freq = 10.0
        w = freq * 2 * pi
        t = linspace(0.0, duration, self.n_samples)

        self.data = sin(w * t)

    def test_amplitude(self):
        """
        Tests the basic algorithm returns the right amplitude
        location.
        """
        amplitude = energy(self.data, self.n_samples)
        max_amp = amax(amplitude)

        ms_sin = 0.5  # The MS energy of a sin wave
        self.assertAlmostEquals(ms_sin, max_amp, places=3)

        # Check that it is in the right location
        self.assertAlmostEqual(max_amp, amplitude[501], places=3)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(EnergyTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




import sys
import glob
import json
import pathlib
import shutil


def hide_cells(notebook, tags=None):
    """
    Finds the tags in each cell and removes it.
    Returns dict without 'hide' tagged cells.
    """
    if tags is None:
        tags = ['hide']

    clean = []
    for cell in notebook['cells']:
        if not set(tags).intersection(cell['metadata'].get('tags', list())):
            clean.append(cell)

    notebook['cells'] = clean

    return notebook


def change_kernel(notebook):
    """
    Vanillafy the kernelspec.
    """
    new_kernelspec = {
        "display_name": "Python 3 (ipykernel)",
        "language": "python",
        "name": "python3",
    }
    notebook['metadata']['kernelspec'].update(new_kernelspec)
    return notebook


def main(path):
    """
    Process the IPYNB files in path, save in place (side-effect).
    """
    fnames = glob.glob(path.strip('/') + '/[!_]*.ipynb')  # Not files with underscore.
    outpath = pathlib.Path('userguide')
    if outpath.exists():
        shutil.rmtree(outpath)
    outpath.mkdir(exist_ok=True)

    for fname in fnames:
        with open(fname, encoding='utf-8') as f:
            notebook = json.loads(f.read())

        notebook = change_kernel(notebook)
        notebook = hide_cells(notebook)
        filepart = pathlib.Path(fname).name

        with open(outpath / filepart, 'w') as f:
            _ = f.write(json.dumps(notebook))

    return


if __name__ == '__main__':
    print(sys.argv[1])
    _ = main(sys.argv[1])


# -*- coding: utf-8 -*-
"""
Tests.
"""
import unittest

from bruges.rockphysics import moduli as m

vp = 2350.
vs = 1125.
rho = 2500
lam = 7478125000.
mu = 3164062500.
youngs = 8551470048.4510355
pr = 0.3513434150638673
pmod = 13806250000.
bulk = 9587500000.


class ModuliTest(unittest.TestCase):
    """
    Tests moduli calculation against spreadsheet
    https://dl.dropboxusercontent.com/u/14965965/Elastic_moduli_formula_checking.xlsx
    """

    def test_vpvs(self):
        self.assertAlmostEqual(m.youngs(vp=vp, vs=vs, rho=rho),
                               youngs, places=-2)
        self.assertAlmostEqual(m.mu(vp=vp, vs=vs, rho=rho), mu, places=-2)
        self.assertAlmostEqual(m.pr(vp=vp, vs=vs, rho=rho), pr)
        self.assertAlmostEqual(m.lam(vp=vp, vs=vs, rho=rho), lam, places=-2)
        self.assertAlmostEqual(m.bulk(vp=vp, vs=vs, rho=rho), bulk, places=-2)
        self.assertAlmostEqual(m.pmod(vp=vp, vs=vs, rho=rho), pmod, places=-2)

    def test_lammu(self):
        self.assertAlmostEqual(m.vp(lam=lam, mu=mu, rho=rho), vp, places=2)
        self.assertAlmostEqual(m.youngs(lam=lam, mu=mu), youngs, places=-2)
        self.assertAlmostEqual(m.pr(lam=lam, mu=mu), pr)
        self.assertAlmostEqual(m.bulk(lam=lam, mu=mu), bulk, places=-2)
        self.assertAlmostEqual(m.pmod(lam=lam, mu=mu), pmod, places=-2)

    def test_youngspr(self):
        self.assertAlmostEqual(m.vp(youngs=youngs, pr=pr, rho=rho),
                               vp, places=2)
        self.assertAlmostEqual(m.mu(youngs=youngs, pr=pr), mu, places=-2)
        self.assertAlmostEqual(m.lam(youngs=youngs, pr=pr), lam, places=-2)
        self.assertAlmostEqual(m.bulk(youngs=youngs, pr=pr), bulk, places=-2)
        self.assertAlmostEqual(m.pmod(youngs=youngs, pr=pr), pmod, places=-2)

    def test_youngslam(self):
        self.assertAlmostEqual(m.mu(youngs=youngs, lam=lam), mu, places=-2)
        self.assertAlmostEqual(m.pr(youngs=youngs, lam=lam), pr, places=-2)
        self.assertAlmostEqual(m.bulk(youngs=youngs, lam=lam), bulk, places=-2)
        self.assertAlmostEqual(m.pmod(youngs=youngs, lam=lam), pmod, places=-2)

    def test_moduli(self):
        mod = {'imp': vp * rho}
        mod['mu'] = mu
        mod['pr'] = pr
        mod['lam'] = lam
        mod['bulk'] = bulk
        mod['pmod'] = pmod
        mod['youngs'] = youngs

        self.assertDictEqual(m.moduli_dict(vp=vp, vs=vs, rho=rho), mod)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(ModuliTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



from .moduli import youngs, bulk, pr
from .moduli import mu, lam, pmod
from .moduli import vp, vs, moduli_dict

from .fluidsub import avseth_gassmann, smith_gassmann
from .fluidsub import vrh, rhogas, rhosat
from .fluidsub import avseth_fluidsub, smith_fluidsub

from .fluids import rho_water, rho_brine
from .fluids import v_water, v_brine
from .fluids import wood

from .anisotropy import backus_parameters, backus
from .anisotropy import backus_quality_factor, thomsen_parameters
from .anisotropy import dispersion_parameter, blangy
from .anisotropy import crack_density
from .anisotropy import hudson_delta_M, hudson_delta_G
from .anisotropy import hudson_quality_factor, hudson_inverse_Q_ratio

from .bounds import voigt_bound, reuss_bound, hill_average, hashin_shtrikman

from .elastic import elastic_impedance

from .rockphysicsmodels import critical_porosity, hertz_mindlin, soft_sand, stiff_sand
from .rockphysicsmodels import contact_cement, constant_cement, increasing_cement
from .rockphysicsmodels import vernik_consol_sand, vernik_sand_diagenesis, vernik_shale
from .rockphysicsmodels import vernik_soft_sand_1, vernik_soft_sand_2



# -*- coding: utf-8 -*-
"""
Tests.
"""
import unittest
import numpy as np

from bruges.transform import time_to_depth, depth_to_time


class TimeDepthTest(unittest.TestCase):
    """
    Tests the basic functionality of the time to depth and depth to
    time conversions in bruges.
    """

    def test_depthtotime(self):
        data = np.zeros((100, 100))
        vmodel = np.zeros((100, 100))

        data[0:50, :] += 100.0
        data[50:, :] += 200.0

        vmodel[0:50, :] += 1500.0
        vmodel[50:, :] += 3000.0

        dt = 0.001
        dz = 1.0

        face_change = int(np.floor(((49 * dz) / 1500.0) / dt))

        output = depth_to_time(data, vmodel, dz, dt, twt=False)

        self.assertTrue((output[face_change+1, 50] -
                         output[face_change, 50]) == 100)

    def test_timetodepth(self):
        data = np.zeros((100, 100))
        vmodel = np.zeros((100, 100))

        data[0:50, :] += 100.0
        data[50:, :] += 200.0

        vmodel[0:50, :] += 1500.0
        vmodel[50:, :] += 5000.0

        dt = 0.001
        dz = 1.0

        output = time_to_depth(data, vmodel, dt, dz, twt=False)

        face_change = int(np.floor(((49.5 * dt) * 1500.0) / dz))

        self.assertTrue((output[face_change+1, 50] -
                         output[face_change, 50]) == 100)

    def test_recip(self):
        data = np.zeros((100, 100))
        vmodel = np.zeros((100, 100))

        data[0:50, :] += 100.0
        data[50:, :] += 200.0

        vmodel[0:50, :] += 1500.0
        vmodel[50:, :] += 1800.0

        dt = 0.001
        dz = 1.

        out1 = depth_to_time(data, vmodel, dz, dt)
        v2 = depth_to_time(vmodel, vmodel, dz, dt)
        out2 = time_to_depth(out1, v2, dt, dz)

        # This test is not finished!


if (__name__ == '__main__'):

    suite = \
      unittest.TestLoader().loadTestsFromTestCase(TimeDepthTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



# -*- coding: utf-8 -*-
"""
===================
petrophysics.py
===================

:copyright: 2018 Agile Geoscience
:license: Apache 2.0
"""

import numpy as np


def gardner(vp, alpha=310, beta=0.25, fps=False):
    """
    Computes Gardner's density prediction from P-wave velocity.

    Args:
        vp (ndarray): P-wave velocity in m/s.
        alpha (float): The factor, 310 for m/s and 230 for ft/s.
        beta (float): The exponent, usually 0.25.
        fps (bool): Set to true for FPS and the equation will use the typical
            value for alpha. Overrides value for alpha, so if you want to use
            you own alpha, regardless of units, set this to False.

    Returns:
        ndarray: RHOB estimate in :math:`kg/m^3`.
    """
    alpha = 230 if fps else alpha
    return alpha * vp ** beta


def inverse_gardner(rho, alpha=310, beta=0.25, fps=False):
    """
    Computes Gardner's density prediction from P-wave velocity.

    Args:
        rho (ndarray): Density in kg/m^3.
        alpha (float): The factor, 310 for m/s and 230 for fps.
        beta (float): The exponent, usually 0.25.
        fps (bool): Set to true for FPS and the equation will use the typical
            value for alpha. Overrides value for alpha, so if you want to use
            your own alpha, regardless of units, set this to False.

    Returns:
        ndarray: Vp estimate in m/s.
    """
    alpha = 230 if fps else alpha
    exponent = 1 / beta
    factor = 1 / alpha**exponent
    return factor * rho**exponent


velocity_to_density = gardner


density_to_velocity = inverse_gardner


def porosity_to_density(phi, rho_matrix, rho_fluid):
    """
    Get density from a porosity log. Typical values:

        - rho_matrix (sandstone) : 2650 kg/m^3
        - rho_matrix (limestome): 2710 kg/m^3
        - rho_matrix (dolomite): 2876 kg/m^3
        - rho_matrix (anyhydrite): 2977 kg/m^3
        - rho_matrix (salt): 20320 kg/m^3

        - rho_fluid (fresh water): 1000 kg/m^3
        - rho_fluid (salt water): 1100 kg/m^3

    See wiki.aapg.org/Density-neutron_log_porosity.

    Args:
        phi (ndarray): The porosity log.
        rho_matrix (float)
        rho_fluid (float)

    Returns:
        Estimate of bulk density, rho.
    """
    return rho_matrix * (1 - phi) + rho_fluid * phi_rhob


def density_to_porosity(rho, rho_matrix, rho_fluid):
    """
    Get density from a porosity log. Typical values:

        - rho_matrix (sandstone) : 2650 kg/m^3
        - rho_matrix (limestome): 2710 kg/m^3
        - rho_matrix (dolomite): 2876 kg/m^3
        - rho_matrix (anyhydrite): 2977 kg/m^3
        - rho_matrix (salt): 20320 kg/m^3

        - rho_fluid (fresh water): 1000 kg/m^3
        - rho_fluid (salt water): 1100 kg/m^3

    See wiki.aapg.org/Density-neutron_log_porosity.

    Args:
        rho (ndarray): The bulk density log or RHOB.
        rho_matrix (float)
        rho_fluid (float)

    Returns:
        Estimate of porosity as a volume fraction.
    """
    return (rho_matrix - rho) / (rho_matrix - rho_fluid)


def slowness_to_velocity(slowness):
    """
    Convert a slowness log in µs per unit depth, to velocity in unit depth
    per second.

    Args:
        slowness (ndarray): A value or sequence of values.

    Returns:
        ndarray: The velocity.
    """
    return 1e6 / np.array(slowness)


velocity_to_slowness = slowness_to_velocity


def gardner_param(vp, rhob):
     """
     Finds optimal alpha and beta parameters for the gardner

     Volodymyr Vragov, October 2018

     Args:
         rho(ndarray): Density.
         vp (ndarray): P-wave velocity.

     Returns:
         alpha (float): The factor.
         beta (float): The exponent, usually 0.25.
     """
     params, _ = curve_fit(optimizer_gardner, rhob, vp)
     return params[0], params[1]
     

def error_flag(pred, actual, dev = 1.0, method = 1):
    """Calculate the difference between a predicted and an actual curve 
    and return a log flagging large differences based on a user-defined distance 
    (in standard deviation units) from the mean difference

    Matteo Niccoli, October 2018
    
    Args:
        predicted (ndarray) = predicted log
        actual (ndarray) =  original log  
        dev  (float) = standard deviations to use, default 1
        error calcluation method (int), default 1
            1: difference between curves larger than mean difference plus dev
            2: curve slopes have opposite sign
            3: curve slopes of opposite sign OR difference larger than mean plus dev
    

    Returns:
    flag (ndarray) =  error flag curve
    """   
    flag = np.zeros(len(pred))
    err = np.abs(pred-actual)
    err_mean = np.mean(err)
    err_std = np.std(err)
    ss = np.sign(pred.diff().fillna(pred))
    ls = np.sign(actual.diff().fillna(actual))

    if method == 1:
        flag[np.where(err>(err_mean + (dev*err_std)))] = 1
    elif method == 2:      
        flag[np.where((ss + ls)==0 )]= 1
    elif method == 3:
        flag[np.where(np.logical_or(err>(err_mean + (dev*err_std)), (ss+ls)==0 ))]= 1
    return flag


def optimize_inverse_gardner(rho, alpha, beta):
     """ Wrapper function to pass inverse_gardner to scipy.curve_fit
     to get optimal alpha and beta parameters

     Matteo Niccoli and Volodymyr Vragov, October 2018

     Args:
         rho(ndarray): Density.
         alpha (float): The factor.
         beta (float): The exponent.

     Returns:
         inverse_gardner: this is passed to scipy.curve_fit as, for example:
         popt_synt, pcov = scipy.curve_fit(optimize_inverse_gardner, rho, vp)
         For a full example, please read:
         mycarta.wordpress.com/2018/10/28/geophysics-python-sprint-2018-day-2-and-beyond-part-i
     """
     return inverse_gardner(rho, alpha=alpha, beta=beta)





"""
Initialize the library.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
from . import attribute
from . import filters
from . import models
from . import petrophysics
from . import rockphysics
from . import reflection
from . import noise
from . import transform
from . import unit
from . import util
from .get_bruges import get_bruges
from .bruges import BrugesError

from pkg_resources import get_distribution, DistributionNotFound

try:
    VERSION = get_distribution(__name__).version
except DistributionNotFound:
    try:
        from ._version import version as VERSION
    except ImportError:
        raise ImportError(
            "Failed to find (autogenerated) _version.py. "
            "This might be because you are installing from GitHub's tarballs, "
            "use the PyPI ones."
            )
__version__ = VERSION



"""
Convolution in n-dimensions.

:copyright: 2022 Agile Geoscience
:license: Apache 2.0
"""
import numpy as np
from bruges.util import apply_along_axis


def convolve(arr, v, axis=-1, verbose=False):
    """
    Convolve n-dimensional arr with a 1D wavelet or 2D wavelet bank.
    
    Args:
        arr (ndarray): The trace, or 2D section, or volume.
        v (ndarray): The wavelet, must be 1D function or a 2D wavelet 'bank'.
            If a wavelet bank, time should be on the last axis.
        axis (int): The time axis of the arr data. In other words, the axis
            corresponding to a single 'trace'. If you index into this axis,
            you will get a single 'trace'.
        verbose (bool): If True, print out the shapes of the inputs and output.

    Returns:
        ndarray: Discrete, linear convolution of `arr` and `v`.
    """
    if v.shape[-1] > arr.shape[axis]:
        raise TypeError("v (e.g. wavelet) must be shorter in time than arr (e.g. reflectivity).")

    arr_ = np.moveaxis(np.asanyarray(arr), axis, -1)

    # Compute the target shape of the final synthetic.
    outshape = v.shape[:-1] + arr_.shape

    # Force v and arr to both be 2D.
    bank = np.atleast_2d(v)   
    arr_2d = arr_.reshape((-1, arr_.shape[-1]))

    # Compute synthetic.
    syn = np.array([apply_along_axis(np.convolve, arr_2d, w, mode='same') for w in bank])

    pos = axis + v.ndim - 1

    out = np.moveaxis(syn.reshape(outshape), -1, pos)

    # Show the shapes of the data we're handling.
    if verbose:
        print(arr.shape, ' * ', v.shape, ' -> ', out.shape)

    return out



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.cm import get_cmap
import io
import base64

from bruges.filters import ricker
from bruges.filters import rotate_phase


def _get_colour(cmap, frac):
    """
    Decide whether to make white or black labels.
    """
    cmap = get_cmap(cmap)
    return 'k' if (np.mean(cmap(frac)[:3]) > 0.5) else 'w'


def _make_synthetic(size=256, top=0.4, base=0.6, value=1, freq=25, phase=0):
    """Make a synthetic. Return the wavelet, the model, the RC, and the synthetic.
    """
    v = np.ones(size) - value

    v[int(top*size):int(base*size)] = value
    rc = np.diff(v)

    w = ricker(0.256, 0.001, freq)

    if phase != 0:
        w = rotate_phase(w, phase, degrees=True)

    syn = np.convolve(rc, w, mode='same')
    return w, v, rc, syn


def polarity_cartoon(layer='hard',
                     polarity='normal',
                     freq='med',
                     phase=0,
                     style='vd',
                     cmap=None,
                     fmt='png',
                     ):
    """
    Plot a polarity cartoon.
    """
    freqs = {'vhi': 60, 'hi': 30, 'med': 15, 'lo': 7.5,
             'vhigh': 60, 'high': 30, 'medium': 15, 'low': 7.5,
             'mod': 15, 'mid': 15}
    backgr = 'soft' if layer == 'hard' else 'hard'
    value = 1 if layer == 'hard' else 0
    size, top, base = 256, 0.4, 0.6

    _, v, _, syn = _make_synthetic(size, top, base, value, freq=freqs[freq], phase=phase)

    if polarity.lower() not in ['normal', 'seg', 'usa', 'us', 'canada']:
        syn *= -1

    if style == 'ramp':
        # cbar is a ramp.
        cbar = np.linspace(-1, 1, size).reshape(-1, 1)
    else:
        # cbar is the synthetic.
        cbar = syn.reshape(-1, 1)

    gs = {'width_ratios':[2,2,2,1]}
    fig, axs = plt.subplots(ncols=4,
                            figsize=(6, 4),
                            gridspec_kw=gs,
                            facecolor='w', sharey=True,
                            )

    # Plot the earth model.
    ax = axs[0]
    cmap_ = 'Greys'
    ax.imshow(v.reshape(-1, 1), aspect='auto', cmap=cmap_, vmin=-1.5, vmax=1.5)
    ax.axhline(top*size, c='w', lw=4)
    ax.axhline(base*size, c='w', lw=4)
    ax.axvline(0.55, c='w', lw=6)  # Total hack to adjust RHS.
    ax.text(0, size/4.75, backgr, ha='center', va='center', color=_get_colour(cmap_, (1-value)*256), size=25)
    ax.text(0, size/2+0.75, layer, ha='center', va='center', color=_get_colour(cmap_, (value)*256), size=25)

    # Plot the impedance diagram.
    ax = axs[1]
    cmap_ = 'Greys'
    ax.imshow(v.reshape(-1, 1), aspect='auto', cmap=cmap_, vmin=0, vmax=2)
    ax.axvline(-0.5, c=(0.58, 0.58, 0.58), lw=50)
    ax.text(0.45, 2*size/8, 'imp', ha='right', va='center', color='k', size=25)
    ax.annotate("", xy=(0.33, size/8), xytext=(0, size/8), arrowprops=dict(arrowstyle="->", connectionstyle="arc3"))

    # Plot the waveform.
    ax = axs[2]
    y = np.arange(syn.size)
    ax.plot(syn, y, 'k')
    ax.fill_betweenx(y, syn, 0, where=syn>0, color='k')
    ax.invert_yaxis()
    ax.text(0.65, size/8, '+', ha='center', va='center', size=30)
    ax.text(-0.65, size/8, '–', ha='center', va='center', size=40)

    # Plot the colourbar.
    ax = axs[3]
    cmap = cmap or 'gray'
    frac = 1/8
    top_col = _get_colour(cmap, frac)
    bot_col = _get_colour(cmap, 7*frac)
    ax.imshow(cbar, cmap=cmap, aspect='auto')
    if style == 'ramp':
        ax.text(0, frac*size, '+', ha='center', va='center', color=top_col, size=30)
        ax.text(0, 7*frac*size, '–', ha='center', va='center', color=bot_col, size=40)

    # Make final adjustments to subplots and figure.
    for ax in axs:
        ax.set_axis_off()

    # Hack.
    plt.subplots_adjust(left=0.1)

    return fig, axs



# -*- coding: utf-8 -*-
"""
Tests.
"""
import unittest

from bruges.rockphysics import bounds as b

# Inputs.
f0 = [0.5, 0.5]
f1 = [50, 50]
f2 = [0.6, 0.4]     # suspension of quartz in water
k1 = [36.0, 4.0]    # Kquartz, Kclay [GPa]
k2 = [36.0, 2.2]    # Kquartz, Kwater [GPa]

# Hashin-Shtrikman tests
fh1 = [0.0, 1.0]     # end-member test case for HS bounds
fh2 = [1.0, 1.0]     # end-member test case for HS bounds
fh3 = [0.4, 0.6]    # middle-ish test case for HS bounds
khs = [21.0, 76.8]  # Bulk moduli [GPa] for HS tests, Clay, Calcite
ghs = [7.0, 32.0]   # Shear moduli [GPa] for HS tests, Clay, Calcite

# Expected outputs.
mv = 20.0       # Voigt average (hand-calculated)
mr = 7.199      # Reuss average (hand-calculated)
mr2 = 5.04      # Example from Mavko et al. RPH
mh = 13.599     # Voigt-Reuss Hill average (hand-calculated)
mhs1 = 76.8     # HS upper and lower bound at end-member 1
mhs2 = 21.0     # HS upper and lower bound at end-member 2
mhs3 = 35.5329  # HS lower bound at 50-50
mhs4 = 40.3989  # HS upper bound at 50-50


class BoundsTest(unittest.TestCase):
    """
    Tests calculations of elastic mixture bounds
    """

    def test_voigt(self):
        self.assertAlmostEqual(b.voigt_bound(f0, k1), mv, places=-2)
        self.assertAlmostEqual(b.voigt_bound(f1, k1), mv, places=-2)

    def test_reuss(self):
        self.assertAlmostEqual(b.reuss_bound(f0, k1), mr, places=-2)
        self.assertAlmostEqual(b.reuss_bound(f2, k2), mr2, places=-2)
        self.assertAlmostEqual(b.reuss_bound([0.6, 0.2, 0.2],
                                             [36.0, 2.2, 0.000131]),
                               0.00065, places=-2)

    def test_hill_average(self):
        self.assertAlmostEqual(b.hill_average(f0, k1), mh, places=-2)

    def test_hashin_shtrikman(self):
        self.assertAlmostEqual(
            b.hashin_shtrikman(fh1, khs, ghs, modulus='bulk')[1],
            mhs1, places=-2)
        self.assertAlmostEqual(
            b.hashin_shtrikman(fh1, khs, ghs, modulus='bulk')[0],
            mhs1, places=-2)
        self.assertAlmostEqual(
            b.hashin_shtrikman(fh2, khs, ghs, modulus='bulk')[1],
            mhs2, places=-2)
        self.assertAlmostEqual(
            b.hashin_shtrikman(fh2, khs, ghs, modulus='bulk')[0],
            mhs2, places=-2)
        self.assertAlmostEqual(
            b.hashin_shtrikman(fh3, khs, ghs, modulus='bulk')[1],
            mhs3, places=-2)
        self.assertAlmostEqual(
            b.hashin_shtrikman(fh3, khs, ghs, modulus='bulk')[0],
            mhs3, places=-2)

if __name__ == '__main__':

    suite = unittest.TestLoader().loadTestsFromTestCase(BoundsTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
Fluid properties.

These functions implement equations from Batzle and Wang (1992), seismic
properties of pore fluids. GEOPHYSICS, VOL. 57, NO. 11; P. 1396-1408,

:copyright: 2018 Agile Geoscience
:license: Apache 2.0
"""
import numpy as np

def wood(Kf1, Kf2, Sf1):
    """
    Wood's equation, per equation 35b in Batzle and Wang (1992).
    """
    return 1 / ((Sf1 / Kf1) + ((1 - Sf1) / Kf2))


def rho_water(temperature, pressure):
    """
    The density of pure water, as a function of temperature and pressure.
    Implements equation 27a from Batzle and Wang (1992).

    Use scalars or arrays; if you use arrays, they must be the same size.

    Args:
        temperature (array): The temperature in degrees Celsius.
        pressure (array): The pressure in pascals.

    Returns:
        array: The density in kg/m3.
    """
    # Align with the symbols and units in Batzle & Wang.
    T, P = np.asanyarray(temperature), np.asanyarray(pressure)*1e-6

    x = -80*T - 3.3*T**2 + 0.00175*T**3 + 489*P - 2*T*P \
        + 0.016*P*T**2 - 1.3e-5*P*T**3 - 0.333*P**2 - 0.002*T*P**2

    return 1000 + 1e-3 * x


def rho_brine(temperature, pressure, salinity):
    """
    The density of NaCl brine, given temperature, pressure, and salinity.
    The density of pure water is computed from rho_water(). Implements
    equation 27b from Batzle and Wang (1992).

    Use scalars or arrays; if you use arrays, they must be the same size.

    Args:
        temperature (array): The temperature in degrees Celsius.
        pressure (array): The pressure in pascals.
        salinity (array): The weight fraction of NaCl, e.g. 35e-3
            for 35 parts per thousand, or 3.5% (the salinity of
            seawater).

    Returns:
        array: The density in kg/m3.
    """
    # Align with the symbols and units in Batzle & Wang.
    T, P = np.asanyarray(temperature), np.asanyarray(pressure)*1e-6
    S = np.asanyarray(salinity)

    rho_w = rho_water(temperature, pressure) / 1000
    x = 300*P - 2400*P*S + T*(80 + 3*T - 3300*S - 13*P + 47*P*S)

    return rho_w + S*(0.668 + 0.44*S + 1e-6 * x)


def rho_gas(temperature, pressure, molecular_weight):
    """
    Gas density, given temperature (in deg C), pressure (in Pa), and molecular
    weight.

    Args:
        temperature (array): The temperature in degrees Celsius.
        pressure (array): The pressure in pascals.
        molecular_weight (array): The molecular weight.
        
    Returns:
        array: The density in kg/m3.
    """
    # Align with the symbols and units in Batzle & Wang.
    T, P = np.asanyarray(temperature), np.asanyarray(pressure)*1e-6
    M = np.asanyarray(molecular_weight)
    R = 8.3144598

    return M * P / (R * (T + 273.15))


def v_water(temperature, pressure):
    """
    The acoustic velocity of pure water, as a function of temperature
    and pressure. Implements equation 28 from Batzle and Wang (1992), using
    the coefficients in Table 1.

    Note that this function does not work at pressures above about 100 MPa.

    Use scalars or arrays; if you use arrays, they must be the same size.

    Args:
        temperature (array): The temperature in degrees Celsius.
        pressure (array): The pressure in pascals.

    Returns:
        array: The velocity in m/s.
    """
    w = np.array([[ 1.40285e+03,  1.52400e+00,  3.43700e-03, -1.19700e-05],
                  [ 4.87100e+00, -1.11000e-02,  1.73900e-04, -1.62800e-06],
                  [-4.78300e-02,  2.74700e-04, -2.13500e-06,  1.23700e-08],
                  [ 1.48700e-04, -6.50300e-07, -1.45500e-08,  1.32700e-10],
                  [-2.19700e-07,  7.98700e-10,  5.23000e-11, -4.61400e-13]])

    T, P = np.asanyarray(temperature), np.asanyarray(pressure) * 1e-6
    return sum(w[i, j] * T**i * P**j for i in range(5) for j in range(4))


def v_brine(temperature, pressure, salinity):
    """
    The acoustic velocity of brine, as a function of temperature (deg C),
    pressure (Pa), and salinity (weight fraction). Implements equation 29
    from Batzle and Wang (1992).

    Note that this function does not work at pressures above about 100 MPa.

    Use scalars or arrays; if you use arrays, they must be the same size.

    Args:
        temperature (array): The temperature in degrees Celsius.
        pressure (array): The pressure in pascals.
        salinity (array): The weight fraction of NaCl, e.g. 35e-3
            for 35 parts per thousand, or 3.5% (the salinity of
            seawater).

    Returns:
        array: The velocity in m/s.
    """
    # Align with the symbols and units in Batzle & Wang.
    T, P = np.asanyarray(temperature), np.asanyarray(pressure)*1e-6
    S = np.asanyarray(salinity)

    v_w = v_water(temperature, pressure)
    s1 = 1170 - 9.6*T + 0.055*T**2 - 8.5e-5*T**3 + 2.6*P - 0.0029*T*P - 0.0476*P**2
    s15 = 780 - 10*P + 0.16*P**2
    s2 = -820

    return v_w + s1 * S + s15 * S**1.5 + s2 * S**2



# -*- coding: utf-8 -*-
"""
Tests.
"""
import unittest

from bruges.rockphysics import fluidsub

# Inputs... GAS case.
vp_gas = 2428.95
vs_gas = 1462.43
rho_gas = 2080.00

# Expected outputs... BRINE case.
vp_brine = 2850.49
vs_brine = 1416.10
rho_brine = 2210.00

# Known parameters.
phi = 0.275           # Don't know this... reading from fig
rhohc = 250.0         # gas, from Fig 5
rhow = 1040.0         # brine, from Fig 5
sw = 0.3              # Don't know this... just guessing
swnew = 1.0           # Don't know this... just guessing
khc = 207000000.0     # gas, from Fig 5
kw = 2950000000.0     # brine, from Fig 5
kclay = 25000000000.0
kqtz = 37000000000.0
vclay = 0.05
kmin = 36266406250.0  # Don't know this... reading from fig


class FluidsubTest(unittest.TestCase):
    """
    Tests fluid sub calculations against Smith et al 2003.
    https://dl.dropboxusercontent.com/u/14965965/Smith_etal_2003.pdf
    """

    def test_avseth(self):
        # Base case: gas
        # Subbing with: brine

        sub = fluidsub.avseth_fluidsub(vp=vp_gas,
                                       vs=vs_gas,
                                       rho=rho_gas,
                                       phi=phi,
                                       rhof1=rhohc,
                                       rhof2=rhow,
                                       kmin=37000000000,
                                       kf1=khc,
                                       kf2=kw)

        self.assertAlmostEqual(sub[0], vp_brine, places=-1)  # Cannot match :(
        self.assertAlmostEqual(sub[1], vs_brine, places=-1)  # Cannot match :(
        self.assertAlmostEqual(sub[2], rho_brine, places=-1)  # Cannot match :(

    def test_smith(self):
        # Base case: gas
        # Subbing with: brine

        sub = fluidsub.smith_fluidsub(vp=vp_gas,
                                      vs=vs_gas,
                                      rho=rho_gas,
                                      phi=phi,
                                      rhohc=rhohc,
                                      rhow=rhow,
                                      sw=sw,
                                      swnew=swnew,
                                      khc=khc,
                                      kw=kw,
                                      kclay=kclay,
                                      kqtz=kqtz,
                                      vclay=vclay)

        self.assertAlmostEqual(sub[0], vp_brine,  places=-1)
        self.assertAlmostEqual(sub[1], vs_brine,  places=-1)
        self.assertAlmostEqual(sub[2], rho_brine, places=-1)  # Cannot match :(


if __name__ == '__main__':

    suite = unittest.TestLoader().loadTestsFromTestCase(FluidsubTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
from .energy import energy

from .discontinuity import discontinuity
from .discontinuity import similarity

from .dipsteer import dipsteer

from .spectrogram import spectrogram

from .spectraldecomp import spectraldecomp

from .complex import instantaneous_amplitude
from .complex import reflection_strength
from .complex import envelope
from .complex import instantaneous_phase
from .complex import instantaneous_frequency
from .complex import quadrature

from .horizon import *



# -*- coding: utf-8 -*-
"""
Coordinate transformation. This module contains a class for
converting between seismic survey inline-xline coordinates
and real-world UTM coordinates.

:copyright: 2018 Agile Geoscience
:license: Apache 2.0
"""
import numpy as np
from ..util.transformations import affine_matrix_from_points


class CoordTransform(object):
    """
    A class for converting between seismic survey inline-xline
    coordinates and real-world UTM coordinates.

    Instantiate with a pair of at least 3 coordinates mapping
    one space to the other. Provide two array-likes of shape
    (3, 2). See below for example.

    After instantiation, the class is callable in the 'forward'
    (inline-xline to UTMx-UTMy) direction. You can also use
    ``CoordTransform.forwrd()``. The ``CoordTransform.reverse()``
    method converts in the other direction (UTMx-UTMy to
    inline-xline).

    Example:
        >>> corner_ix = [[0,  0], [0, 950], [650, 950]]
        >>> corner_xy = [[605835.5, 6073556.5],
                        [629576.3, 6074220.0],
                        [629122.5, 6090463.2]]
        >>> transform = bruges.transform.CoordTransform(corner_ix, corner_xy)
        >>> transform([300, 400])
        array([ 615622.18016194, 6081332.72995951])
        >>> transform.forward([300, 400])
        array([ 615622.18016194, 6081332.72995951])
        >>> transform.reverse([ 615622.18016194, 6081332.72995951])
        ​array([300, 400])
    """
    def __init__(self, ix, xy):
        ix = np.array(ix)
        xy = np.array(xy)
        if ix.shape[1] == 2:
            ix = ix[:3].T
        if xy.shape[1] == 2:
            xy = xy[:3].T
        self.A = affine_matrix_from_points(ix, xy)


    def __call__(self, p):
        p = np.asanyarray(p)
        return np.dot(self.A, np.append(p, [1]))[:2]


    def forward(self, p):
        """
        Convert inline-xline to UTMx-UTM-y.

        Example:
            >>> transform.forward([300, 400])
            array([ 615622.18016194, 6081332.72995951])
        """
        p = np.asanyarray(p)
        return self(p)


    def reverse(self, q):
        """
        Convert UTMx-UTM-y to inline-xline.

        Example:
            >>> transform.reverse([ 615622.18016194, 6081332.72995951])
        ​   array([300, 400])
        """
        p = np.dot(np.linalg.pinv(self.A), np.append(q, [1]))[:2]
        return np.rint(p).astype(np.int)



import unittest
import numpy as np

from bruges.attribute import dipsteer
from bruges.filters import ricker


class DipTest(unittest.TestCase):

    # Parking this failing test; I'm not convinced that the dipsteer
    # function is working correctly.
    #
    # def test_spikes(self):
    #     """
    #     Make a simple case with spikes offset on each trace.
    #     """

    #     test_data = np.zeros((100, 100))
    #     index = np.arange(0, 100)

    #     test_data[index, index] = 1.0

    #     maxlag = 5
    #     dip = dipsteer(test_data, 1, 1, maxlag)

    #     # Output should be all -1 except for the edge effects
    #     same = np.allclose(dip[maxlag:-maxlag, maxlag:-maxlag],
    #                        -test_data[maxlag:-maxlag, maxlag:-maxlag],
    #                        .01)

    #     self.assertTrue(same)

    #     # Answer should not change if we increase the stepout
    #     stepout = 4
    #     maxlag = 1
    #     dip = dipsteer(test_data, 1, stepout, maxlag)

    #     diff = (dip[maxlag*stepout:100-(maxlag*stepout),
    #             stepout * maxlag:100-(stepout*maxlag)] +
    #             test_data[maxlag * stepout:100-(maxlag * stepout),
    #                       maxlag*stepout:100-(stepout*maxlag)])

    #     diff = diff[1:-1, 1:-1]

    #     # Output should be all -1 except for the edge effects
    #     same = np.allclose(diff, np.zeros(diff.shape), .01)

    #     self.assertTrue(same)

    def test_wavelet(self):

        # Make a dipping reflectivity field
        data = np.zeros((150, 100))

        data[0, :] = np.arange(100) / 10.0

        dips = np.arange(0, 100, 2)
        dips = np.concatenate([dips, dips[::-1]])

        window_size = 10
        freq = 20.0
        duration = 1
        wavelet, _ = ricker(window_size, duration, freq)

        for i in range(dips.size):

            data[:, i] = np.convolve(wavelet,
                                     np.roll(data[:, i], dips[i]),
                                     mode='same')

        stepout = 1
        maxlag = 2
        dip = dipsteer(data, window_size, stepout=stepout,
                       maxlag=maxlag, overlap=0.5)

        undipped = np.zeros(dips.shape)
        for i in range(dips.size):
            undipped[i] = (dip[dips[i], i])

        check = undipped[maxlag*stepout:-maxlag*stepout]
        test = np.zeros(check.shape)
        test[0: test.size//2 - 1] = -2.0
        test[(test.size//2) - 1] = -1
        test[(test.size//2)] = 1
        test[(test.size//2)+1:] = 2

        self.assertTrue(np.allclose(test, check))


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(DipTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



# -*- coding: utf-8 -*-
"""
Testing util.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
import unittest
import numpy as np
from bruges.util import next_pow2, rms, top_and_tail


class UtilTest(unittest.TestCase):

    def test_toptail(self):
        a = np.asarray([np.nan, np.nan, 3, 4, 5, np.nan])
        b = np.asarray([6, 5, 4, 3, 2, 1])
        c = np.asarray([1, 2, 3, 4, 5, 6])
        d = np.asarray([6, 5, 4, 3, 2, 1])
        A, B, C, D = top_and_tail(a, b, c, d)
        for arr in (A, B, C, D):
            self.assertEqual(len(arr), 3)

    def test_nextpow2(self):
        num = 888
        ans = 1024

        self.assertEqual(ans, next_pow2(num))

    def test_rms(self):
        l = np.array([2, 3, 4.5])
        ans = 3.329164
        self.assertAlmostEqual(ans, rms(l), places=5)

        l = [[[1, 2, 3], [4, 5, 6]], [[2,3,4], [5, 6, 7]]]
        ans = 2.1602469
        self.assertAlmostEqual(ans, rms(l, axis=-1)[0, 0], places=5)

    def noise_db(self):
        a = np.ones(1000000)
        ans10p = 3.32
        ans0 = 1.41
        ans10n = 1.05

        self.assertAlmostEqual(ans10p, noise_db(a, 10), places=2)
        self.assertAlmostEqual(ans0, noise_db(a, 10), places=2)
        self.assertAlmostEqual(ans10n, noise_db(a, -10), places=2)

if __name__ == '__main__':

    suite = unittest.TestLoader().loadTestsFromTestCase(UtilityTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



# -*- coding: utf-8 -*-
import unittest
import numpy

from bruges.attribute import similarity


class SimilarityTest(unittest.TestCase):

    def test_same_data(self):
        """
        Simple test to check if the algorithm works for the trivial case.
        """
        # data2d = numpy.ones([3, 100])
        # output2d = similarity(data2d, duration=0.01, dt=0.001, kind='marfurt')
        # self.assertTrue(output2d.dims == 2)

        data3d = numpy.ones([3, 3, 100])
        output3d = similarity(data3d, duration=0.01, dt=0.001, kind='marfurt')
        self.assertTrue(output3d.ndim == 3)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(SimilarityTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



import unittest
import numpy as np
from numpy import array

from bruges.models import reconcile, interpolate, panel
from bruges.models import wedge


class ModelTest(unittest.TestCase):
    """
    Tests models.
    """
    def test_reconcile(self):
        a = np.array([2, 6, 7, 7, 3])
        b = np.array([3, 7, 3])
        A, B = reconcile(a, b, order=0)
        A_, B_ = array([2, 6, 7, 7, 3]), array([3, 7, 7, 3, 3])
        self.assertTrue(np.array_equal(A, A_))
        self.assertTrue(np.array_equal(B, B_))

    def test_interpolate(self):
        a = np.array([2, 6, 7, 7, 3])
        b = np.array([3, 7, 7, 3, 3])
        interp = interpolate(a, b, num=10)
        self.assertTrue(interp.shape == (5, 10))

    def test_panel(self):
        a = np.array([2, 6, 7, 7, 3])
        b = np.array([3, 7, 3])
        dists = (10,)
        out = panel(a, b, num=15, dists=dists)
        sample = out[:, 7]
        self.assertTrue(np.all(sample[:4] == array([2.5, 6.5, 5., 3.])))
        self.assertTrue(np.isnan(sample[-1]))

    def test_wedge(self):
        w, top, base, ref = wedge(depth=10, width=7, strat=(10, (20, 30), 40))
        col = array([10, 10, 10, 20, 20, 30, 40, 40, 40, 40])
        t = array([3., 3., 3., 3., 3., 3., 3.])
        b = array([3., 3., 3.6, 4.2, 4.8, 5.4, 6. ])
        self.assertTrue(np.all(w[:, -1] == col))
        self.assertTrue(w.sum() == 1990)
        self.assertTrue(np.allclose(top, t))
        self.assertTrue(np.allclose(base, b))
        self.assertTrue(ref == 6)

    def test_wedge_numpy(self):
        strat = np.array([2000, 2100, 2300])
        w, *_ = wedge(depth=10, width=7, strat=strat)
        self.assertTrue(w.shape == (10, 7))

    def test_netgross(self):
        w, top, *_ = wedge(depth=10, width=7, breadth=3, strat=(10, (20, 30), 40))
        self.assertTrue(w.sum() == 6003)
        self.assertTrue(w.shape == (10, 7, 3))
        self.assertTrue(top.sum() == 63.0)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(ModelTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



from collections import namedtuple

import numpy as np
import scipy.ndimage as sn
import scipy.ndimage.morphology as morph

from bruges.util import sigmoid, root, power


def pad_func(before, after):
    """
    Padding function. Operates on vector *in place*, per the np.pad
    documentation.
    """
    def pad_with(x, pad_width, iaxis, kwargs):
        x[:pad_width[0]] = before[-pad_width[0]:]
        x[-pad_width[1]:] = after[:pad_width[1]]
        return
    return pad_with


def get_strat(strat,
              thickness,
              kind='nearest',
              position=1,
              wedge=None,
              zoom_mode='nearest'
              ):
    """
    Take a 'stratigraphy' (either an int, a tuple of ints, or a list-like of
    floats) and expand or compress it to the required thickness.

    `kind` can be 'nearest', 'linear', 'quadratic', or 'cubic'.
    """
    orders = {'nearest': 0, 'linear': 1, 'quadratic': 2, 'cubic': 3}
    order = orders.get(kind, 0)

    if np.issubdtype(type(strat), int) and (order == 0):
        out = np.repeat([strat], thickness)
    elif np.issubdtype(type(strat), float) and (order == 0):
        out = np.repeat([strat], thickness)
    elif isinstance(strat, tuple) and (order == 0):
        out = np.repeat(strat, int(round(thickness/len(strat))))
    else:
        if position == 0:
            wedge_zoom = wedge[1]/len(wedge[0])
            strat = strat[-int(thickness/wedge_zoom):]
        elif position == -1:
            wedge_zoom = wedge[1]/len(wedge[0])
            strat = strat[:int(thickness/wedge_zoom)]
        zoom = thickness / len(strat)
        out = sn.zoom(strat, zoom=zoom, order=order, mode=zoom_mode)

    # Guarantee correct length by adjusting bottom layer.
    missing = int(np.ceil(thickness - out.size))
    if out.size > 0 and missing > 0:
        out = np.pad(out, [0, missing], mode='edge')
    elif out.size > 0 and missing < 0:
        out = out[:missing]

    return out


def get_conforming(strat, thickness, conformance):
    """
    Function to deal with top and bottom conforming wedges.
    """
    thickness = int(np.ceil(thickness))
    if thickness == 0:
        return np.array([])
    if strat.size == thickness:
        return strat
    elif strat.size > thickness:
        if conformance == 'top':
            return strat[:thickness]
        else:
            return strat[-thickness:]
    else:
        if conformance == 'top':
            return np.pad(strat, [0, thickness-strat.size], mode='wrap')
        else:
            return np.pad(strat, [thickness-strat.size, 0], mode='wrap')
    return


def get_subwedges(target, breadth):
    """
    For a binary target (the reference trace of a wedge),
    create the range of net:gross subwedges. We do this
    with binary morphologies in the following way:

    - Erode the 'secondary' component (whatever appears
      second in the target array) one step at a time,
      until there is nothing left and the resulting
      trace contains only the primary.
    - Dilate the secondary component until there is
      nothing left of the primary and the resulting trace
      contains only the secondary.
    - Arrange the traces in order, starting with all
      primary, and ending in all secondary. The target
      trace will be somewhere in between, but not
      necessarily in the middle.

    Returns a 2D array with one target wedge trace per
    section.

    Args:
        target (array): A 1D array length N, the 'reference'
            trace for the wedge. The reference trace has
            thickness '1' in the wedge model. This trace must
            be 'binary' — i.e. it must contain exactly 2
            unique values.
        breadth (int): How many new reference traces should
            be in the output.

    Returns:
        tuple (ndarray, ndarray, int). The ndarray has shape
            N x breadth. It represents one target wedge trace
            per section in 'breadth'. The integer is the
            position of the target trace in the ndarray's
            second dimension.
    """
    try:
        components = a, b = np.unique(target)
    except ValueError:
        raise ValueError("Must be a binary (2-component) wedge.")

    out = [target]

    temp = target.copy()
    while b in temp:
        ero = morph.binary_erosion(temp == b)
        temp = components[ero.astype(int)]
        out.append(temp)

    out = out[::-1]
    ref = len(out)

    temp = target.copy()
    while a in temp:
        dil = morph.binary_dilation(temp == b)
        temp = components[dil.astype(int)]
        out.append(temp)

    arr_ = np.array(out).T
    h, w = arr_.shape
    arr = sn.zoom(arr_, zoom=(1, breadth/w), mode='nearest')

    ng = np.divide(np.sum(arr == a, axis=0), target.size)

    return arr, ng, ref * breadth / w


def wedge(depth=(30, 40, 30),
          width=(10, 80, 10),
          breadth=None,
          strat=(0, 1, 2),
          thickness=(0.0, 1.0),
          mode='linear',
          conformance='both',
          ):
    """
    Generate a wedge model.

    Args:
        depth (int or tuple): The vertical size of the model. If a 3-tuple,
            then each element corresponds to a layer. If an integer, then each
            layer of the model will be 1/3 of the thickness. Note that if the
            'right' wedge thickness is more than 1, then the total thickness
            will be greater than this value.
        width (int or tuple): The width of the model. If a 3-tuple, then each
            element corresponds to a 'zone' (left, middle, right). If an
            integer, then the zones will be 10%, 80% and 10% of the width,
            respectively.
        breadth (None or int): Not implemented. Raises an error.
        strat (tuple): Stratigraphy above, in, and below the wedge. This is the
            'reference' stratigraphy. If you give integers, you get 'solid'
            layers containing those numbers. If you give arrays, you will get
            layers of those arrays, expanded or squeezed into the layer
            thicknesses implied in `depth`.
        thickness (tuple): The wedge thickness on the left and on the right.
            Default is (0.0, 1.0) so the wedge will be thickness 0 on the left
            and the wedge thickness specified in the depth argument on the
            right. If the thickness are equal, you'll have a flat, layer-cake
            model.
        mode (str or function): What kind of interpolation to use. Default:
            'linear'. Other options are 'sigmoid', which makes a clinoform-like
            body, 'root', which makes a steep-sided bowl shape like the edge of
            a channel, and 'power', which makes a less steep-sided bowl shape.
            If you pass a function, give it an API like np.linspace: f(start,
            stop, num), where start is the left thickness, stop is the right
            thickness, and num is the width of the middle (wedge) 'zone'.
        conformance (str): 'top', 'bottom', or 'both' (the default). How you
            want the layers inside the wedge to behave. For top and bottom
            conformance, if the layer needs to be thicker than the reference.

    Returns:
        namedtuple[ndarray, ndarray, ndarray, int]: A tuple containing the
            2D wedge model, the top 'horizon', the base 'horizon', and the
            position at which the wedge has thickness 1 (i.e. is the thickness
            specfied by the middle layer depth and/or strat).
    """
    # Decide if binary (2 rocks in the wedge).
    if np.unique(strat[1]).size == 2:
        binary = True
    else:
        binary = False

    if breadth is None or breadth == 0:
        breadth = 1

    # Allow wedge to be thin-thick or thick-thin.
    left, right = thickness
    if left > right:
        left, right = right, left
        flip = True
    else:
        flip = False

    # Get all layers thicknesses.
    if isinstance(depth, int):
        L1, L2, _ = 3 * [depth//3]  # Sizes if depth is just a number.
        L3 = depth - L1 - L2
    else:
        L1, L2, L3 = map(int, depth)
    L3 += int(right * L2)  # Adjust bottom layer.

    # Get all zone widths.
    if isinstance(width, int):
        Z2 = round(0.8 * width)
        Z1 = round(max(1, width/10))
        Z3 = width - Z2 - Z1
        width = int(Z1), int(Z2), int(Z3)
    else:
        Z1, Z2, Z3 = width  # Z1 and Z3 are the bookends.

    # Deal with different interpolation patterns.
    modes = {
        'linear': np.linspace,
        'clinoform': sigmoid,
        'sigmoid': sigmoid,
        'root': root,
        'power': power,
    }
    zooms = modes.get(mode, mode)(left, right, Z2)

    # Get the reference stratigraphy in each layer.
    # The 'well log' case is tricky, because layer1 and layer3
    # need to know about the amount of zoom on the wedge layer.
    # There must be an easier way to do this.
    layer1 = get_strat(strat[0], L1, position=0, wedge=(strat[1], L2))
    layer2_ = get_strat(strat[1], L2, position=1)
    layer3 = get_strat(strat[2], L3, position=-1, wedge=(strat[1], L2))

    # Deal with width. We discard the reference breadth.
    if binary and (breadth >= 2):
        layer2s, _n2g, _ = get_subwedges(layer2_, breadth=breadth)
        layer2s = layer2s.T
    else:
        layer2s, _ = [layer2_], None

    # Make the padding function.
    padder = pad_func(layer1, layer3)

    # For everything in breadth:
    model = []
    for layer2 in layer2s:
        # Collect wedge pieces, then pad top & bottom, then stack,
        # then pad left & right.
        if conformance in ['top', 'bottom', 'base']:
            wedges = [get_conforming(layer2, z*L2, conformance) for z in zooms]
        else:
            wedges = [get_strat(layer2, thickness=z*L2) for z in zooms]
        padded = [np.pad(w, [L1, L3-w.size], mode=padder) for w in wedges]
        wedge = np.pad(np.stack(padded), [[Z1, Z3], [0, 0]], mode='edge')
        model.append(wedge.T)
    model = np.array(model)

    # Make the top and base 'horizons'.
    top = np.repeat((np.ones(np.sum(width)) * L1)[:, None], breadth, axis=-1)
    base_ = np.pad(L1 + zooms * L2, [Z1, Z3], mode='edge')
    base = np.repeat(base_[:, None], breadth, axis=-1)

    # Calculate the reference profile ('well' position).
    if left <= 1 <= right:
        ref = Z1 + np.argmin(np.abs(zooms-1))
    elif left == right == 1:
        ref = Z1 + Z2//2
    else:
        ref = -1

    # Flip if required.
    if flip:
        model = np.flip(model, axis=2)
        base = base[::-1]
        ref = sum(width) - ref

    # Move the 'breadth' dim to last.
    if model.shape[0] > 1:
        model = np.moveaxis(model, 0, -1)

    # Fix any shape discrepancy.
    # if model.shape

    # Build and return output.
    Wedge = namedtuple('Wedge', ['wedge', 'top', 'base', 'reference'])
    return Wedge(np.squeeze(model),
                 np.squeeze(top),
                 np.squeeze(base),
                 ref
                 )



"""
Various methods to compute seismic discontinuity. 

:copyright: 2015 Joe Kington, SEG Tutorial https://library.seg.org/doi/pdf/10.1190/tle34121510.1
:license: Apache 2.0
"""
import numpy as np
from scipy.ndimage import generic_filter
from scipy.ndimage import gaussian_filter1d


def moving_window(traces, func, window):
    """
    Helper function for multi-trace attribute generation.
    This function applies a 3D function func to process a
    region of shape `window` over a dataset `data`.
    """
    wrapped = lambda x: func(x.reshape(window))
    return generic_filter(traces, wrapped, window)


def marfurt(traces):
    """
    Marfurt, K., V. Sudhaker, A. Gersztenkorn, K. D. Crawford, and S. E. Nissen, 1999,
    Coherency calculations in the presence of structural dip: GEOPHYSICS, 64, 104-111.
    doi:10.1190/1.1444508
    """
    i, x, t = traces.shape
    traces = traces.reshape(-1, t)
    square_sums = np.sum(traces, axis=0)**2
    sum_squares = np.sum(traces**2, axis=0)
    c = square_sums.sum() / (sum_squares.sum() + 1e-12)
    return c / (i * x)


def gersztenkorn(traces):
    """
    Gersztenkorn, A., and K. J. Marfurt, 1999, Eigenstructure‐based coherence
    computations as an aid to 3-D structural and stratigraphic mapping:
    GEOPHYSICS, 64, 1468-1479. doi:10.1190/1.1444651
    """
    # Stack traces in 3D traces into 2D array.
    traces = traces.reshape(-1, traces.shape[-1])

    # Calculate eigenvalues of covariance matrix.
    cov = traces.dot(traces.T)
    vals = np.linalg.eigvalsh(cov)
    return vals.max() / (vals.sum() + 1e-12)


def gradients(traces, sigma):
    grads = []
    for axis in range(3):
        grad = gaussian_filter1d(traces, sigma, axis=axis, order=1)
        grads.append(grad[..., np.newaxis])
    return np.concatenate(grads, axis=3)


def moving_window4d(grad, window, func):
    """Applies the given function *func* over a moving *window*, reducing 
    the input *grad* array from 4D to 3D."""
    # Pad in the spatial dimensions, but leave the gradient dimension unpadded.
    half_window = [(x // 2, x // 2) for x in window] + [(0, 0)]
    padded = np.pad(grad, half_window, mode='reflect')
    
    out = np.empty(grad.shape[:3], dtype=float)
    for i, j, k in np.ndindex(out.shape):
        region = padded[i:i+window[0], j:j+window[1], k:k+window[2], :]
        out[i,j,k] = func(region)
    return out


def gst_calc(region):
    region = region.reshape(-1, 3)
    gst = region.T.dot(region)
    eigs = np.sort(np.linalg.eigvalsh(gst))[::-1]
    return (eigs[0]-eigs[1]) / (eigs[0]+eigs[1])


def gst_discontinuity(seismic, window, sigma=1):
    """
    Gradient structure tensor discontinuity.

    Randen, T., E. Monsen, C. Singe, A. Abrahamsen, J. Hansen, T. Saeter, and J. Schlaf, 2000,
    Three-dimensional texture attributes for seismic data analysis, 70th Annual International Meeting,
    SEG, Expanded Abstracts, 668-671.
    """
    grad = gradients(seismic, sigma)
    return moving_window4d(grad, window, gst_calc)


def discontinuity(traces, duration, dt, step_out=1, kind='gst', sigma=1):
    """
    Compute discontinuity for a seismic section using one of various methods.

    Expects time or depth to be in the last axis of a 2D or 3D input.

    :param traces: A 2D or 3D NumPy array arranged as (cdp, twt) or
        (iline, xline, twt).
    :param duration: The length in seconds of the window trace kernel
        used to calculate the discontinuity.
    :keyword dt (default=1): The sample interval of the traces in sec.
        (eg. 0.001, 0.002, ...). Will default to one, allowing
        duration to be given in samples.
    :keyword step_out (default=1):
        The number of adjacent traces to the kernel to compute discontinuity over.
    :keyword kind (default='gst'):
        The method to use for the computation. Can be "marfurt", "gersztenkorn"
        or "gst" (gradient structure tensor).
    :keyword sigma (default=1):
        The width of the Gaussian function used to compute gradients.
    """
    if traces.ndim == 2:
        traces = traces[:, None, :]
        window = 2*step_out+1, 1, int(duration / dt)
    elif traces.ndim == 3:
        window = 2*step_out+1, int(duration / dt), 2*step_out+1
    else:
        raise NotImplementedError("Expected 2D or 3D seismic data.")

    methods = {
        "marfurt": moving_window(traces, marfurt, window),
        "gersztenkorn": moving_window(traces, gersztenkorn, window),
        "gst": gst_discontinuity(traces, window, sigma)
    }

    return np.squeeze(methods[kind])

similarity = discontinuity


# -*- coding: utf-8 -*-
"""
Spectrogram.

:copyright: 2015 Agile Geoscience
:license: Apache 2.0
"""
import numpy as np
from scipy.fftpack import fft
from scipy.signal import get_window

from bruges.util import next_pow2


def spectrogram(data, window_length,
                dt=1.0,
                window_type='boxcar',
                overlap=0.5,
                normalize=False,
                zero_padding=0):
    """
    Calculates a spectrogram using windowed STFTs.

    Args:
        data (array): 1D numpy array to process into spectra.
    
        window_length (float): The length of the window in seconds if dt is set, otherwise in samples. Will zero pad to the closest power  of two.
    
        window_type: A string specifying the type of window to use for the STFT. The same input as scipy.signal.get_window
    
        dt (float or int): The time sample interval of the trace. Defaults to 1, which allows window_length to be in seconds.
    
        overlap (float): The fractional overlap for each window. A value of 0 uses no redudant data, a value of 1 slides the STFT window one sample at a time. Defaults to 0.5
        
        normalize (bool): Normalizes the each spectral slice to have unity energy.
        
        zero_padding (float or int): The amount of zero padding to when creating the spectra.

    Returns
    -------
    out : A spectrogram of the data ([time, freq]). (2D array for 1D input)
    
    See Also
    --------
    spectraldecomp : Spectral decomposition
    
    """
    # Make the base window
    window_n = int(np.floor(window_length / dt))
    pad = int(np.floor(zero_padding / dt))
    window = get_window(window_type, window_n)

    # Calculate how many STFTs we need to do.
    stride = int(window.size * (1 - overlap) + 1)
    n_windows = int(np.ceil((data.size - window.size) / stride) + 1)

    # Pad window to power of 2
    padded_window = np.zeros(next_pow2(window.size+pad))

    # Init the output array
    output = np.zeros([n_windows, int(padded_window.size // 2)])

    # Do the loop
    for i in range(int(n_windows)):

        start = int(i * stride)
        end = start + int(window.size)

        # Do not compute last samples if we don't have a full window
        if (end > data.size-1):
            break

        padded_window[0:window.size] = window*data[start:end]
        spect = (2. * np.absolute(fft(padded_window)) /
                 window.size)[0:int(padded_window.size // 2)]

        if normalize:
            output[i, :] = spect / np.sum(spect**2)

        else:
            output[i, :] = spect

    return output



import unittest
import numpy as np

from bruges.filters import convolve


class ConvolveTest(unittest.TestCase):

    boxcar = np.array([1, 1, 1, 1, 1, 1, 1]) / 7
    sinc = np.array([-0.        ,  0.        ,  0.        ,  0.        , -0.15321277,
                  -0.1887466 ,  0.17251039,  0.72943393,  1.        ,  0.72943393,
                   0.17251039, -0.1887466 , -0.15321277,  0.        ,  0.        ,
                   0.        , -0.        ])

    def test_1d(self):
        rng = np.random.default_rng(42)
        data = rng.uniform(size=96)
        result = convolve(data, self.boxcar)
        self.assertTrue(result.shape == data.shape)

        data2 = np.zeros(100)
        data2[50] = 1
        result2 = convolve(data2, self.boxcar)
        self.assertTrue(result2.sum() - 7 < 1e-6)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(ConvolveTest)
    unittest.TextTestRunner(verbosity=2).run(suite)
